#! /usr/bin/env python

"""
Drivers to perform global optimization.

Global optimizations can be performed sequentially on a local machine using
:class:`~gc3libs.optimizer.drivers.SequentialDriver`. To make use of
parallelization, :class:`~gc3libs.optimizer.drivers.ParallelDriver` allows
submission of jobs to gc3pie ressources.

Drivers use an algorithm instance that conforms to
:class:`optimizer.EvolutionaryAlgorithm <gc3libs.optimizer.EvolutionaryAlgorithm>` to generate new
populations.
"""

# Copyright (C) 2011, 2012, 2013  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
from builtins import str
from builtins import range
from builtins import object
__author__ = 'Benjamin Jonen <benjamin.jonen@bf.uzh.ch>'
__docformat__ = 'reStructuredText'

import os
import logging
import datetime

import numpy as np
from prettytable import PrettyTable

import gc3libs
from gc3libs.workflow import SequentialTaskCollection, ParallelTaskCollection


class SequentialDriver(object):

    """Drives an optimization using `opt_algorithm` on the local machine.

    The user-supplied :func:`target_fun` computes target values for the
    populations generated by `opt_algorithm`.

    :param opt_algorithm: Evolutionary algorithm instance that conforms to
                          :class:`optimizer.EvolutionaryAlgorithm
                          <gc3libs.optimizer.EvolutionaryAlgorithm>`.

    :param target_fn: Function to evaluate a population and return the corresponding values.
    :param path_to_stage_dir: Directory in which to perform the optimization.
    :param cur_pop_file: Filename under which the population is stored in the
                         current iteration dir. The population is discarded
                         if no file is specified.
    :param logger: Configured logger to use.

    :param str fmt: ``%``-format string to use (e.g., ``%12.8f``) to print
                     values at each step of the algorithm. If `None` (default), this verbose
                     report is not generated, as it might be time-consuming for large population
                     sizes.
    """

    def __init__(
            self,
            opt_algorithm,
            target_fn,
            path_to_stage_dir=os.getcwd(),
            cur_pop_file=None,
            logger=None,
            fmt=None):
        self.path_to_stage_dir = path_to_stage_dir
        self.opt_algorithm = opt_algorithm
        self.target_fn = target_fn
        self.cur_pop_file = cur_pop_file
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger('gc3.gc3libs')
        self.fmt = fmt

    def de_opt(self):
        '''
        Drives optimization until convergence or `itermax` is reached.
        '''
        self.logger.debug('entering de_opt')
        new_pop = self.opt_algorithm.pop
        has_converged = False
        while not has_converged and self.opt_algorithm.cur_iter <= self.opt_algorithm.itermax:
            # Save current population
            self.iteration_folder = os.path.join(
                self.path_to_stage_dir, 'iter_' + str(self.opt_algorithm.cur_iter))
            if self.cur_pop_file:
                if not os.path.isdir(self.iteration_folder):
                    os.mkdir(self.iteration_folder)
                np.savetxt(
                    os.path.join(
                        self.iteration_folder,
                        self.cur_pop_file),
                    new_pop,
                    delimiter=' ')
            # EVALUATE TARGET #
            new_vals = self.target_fn(new_pop)
            if self.fmt:
                self.logger.info(
                    "*** Population (X's) and values (Y) at iteration %d: ***",
                    self.opt_algorithm.cur_iter)
                tb = PrettyTable()
                for i in range(new_pop.shape[1]):
                    tb.add_column(
                        ("X_%d" %
                         i), [
                            (self.fmt %
                             x) for x in new_pop[
                                :, i]])
                tb.add_column("Y", [(self.fmt % y) for y in new_vals])
                for line in tb.get_string().split('\n'):
                    self.logger.info(line)
            # if __debug__:
                #self.logger.debug('x -> f(x)')
                # for x, fx in zip(new_pop, new_vals):
                    #self.logger.debug('%s -> %s' % (x.tolist(), fx))
            self.opt_algorithm.update_opt_state(new_pop, new_vals)
            # create output
            has_converged = self.opt_algorithm.has_converged()
            new_pop = self.opt_algorithm.evolve()
        self.logger.debug('exiting ' + __name__)


class ParallelDriver(SequentialTaskCollection):

    """Drives an optimization using `opt_algorithm` on the grid.

    At each iteration an instance of :class:`ComputeTargetVals` uses
    :func:`task_constructor` to generate :class:`gc3libs.Application` instances
    to be executed in parallel. When all jobs are complete, the output is
    analyzed with the user-supplied function :func:`extract_value_fn`. This
    function returns the function value for all analyzed input vectors.

    :param str jobname:       string that labels this optimization case.

    :param path_to_stage_dir: directory in which to perform the optimization.

    :param opt_algorithm: Evolutionary algorithm instance that conforms to
                          :class:`optimizer.EvolutionaryAlgorithm
                          <gc3libs.optimizer.EvolutionaryAlgorithm>`.

    :param task_constructor: A function that takes a list of x vectors
                             and the path to the current iteration
                             directory, and returns Application
                             instances that can be executed on the
                             grid.

    :param extract_value_fn: Takes an `Application`:class: instance returns
                             the function value computed in that task.
                             The default implementation just looks for a
                             `.value` attribute on the application instance.

    :param cur_pop_file:     Filename under which the population is stored
                             in the current iteration dir. The
                             population is discarded if no file is
                             specified.


    Optimization drivers use GC3Pie in the following way: A
    :class:`~gc3libs.workflow.SequentialTaskCollection` represents the main
    loop of the optimization algorithm, checking for convergence at each
    iteration. This allows for resuming paused or crashed optimizations. Each
    iteration, the optimization algorithm provides a new set of points to be
    evaluated. These points are each represented by an
    :class:`~gc3libs.Application` and bundled into a
    :class:`~gc3libs.workflow.ParallelTaskCollection` that manages each single
    :class:`~gc3libs.Application` until completion. The structure of GC3Libs
    objects employed can be summarized as follows::

         SequentialTaskCollection
                   |
                   v
           ParallelTaskCollection
                  |
                  v
             Application


    """

    def __init__(self, jobname='', path_to_stage_dir='',
                 opt_algorithm=None, task_constructor=None,
                 extract_value_fn=(lambda app: app.value),
                 cur_pop_file = '',
                 **extra_args):

        gc3libs.log.debug('entering ParallelDriver.__init__')

        # Set up initial variables and set the correct methods.
        self.jobname = jobname
        self.path_to_stage_dir = path_to_stage_dir
        self.opt_algorithm = opt_algorithm
        self.extract_value_fn = extract_value_fn
        self.task_constructor = task_constructor
        self.cur_pop_file = cur_pop_file
        self.extra_args = extra_args

        self.new_pop = self.opt_algorithm.pop
        initial_task = ComputeTargetVals(
            self.opt_algorithm.pop, self.jobname, self.opt_algorithm.cur_iter,
            path_to_stage_dir, self.cur_pop_file, task_constructor)

        SequentialTaskCollection.__init__(self, [initial_task], **extra_args)

    def next(self, done):
        gc3libs.log.debug('entering ParallelDriver.next(%d)', done)

        # feed back results from the evaluation just completed
        new_pop = self.new_pop
        new_vals = [self.extract_value_fn(task)
                    for task in self.tasks[done].tasks]
        self.opt_algorithm.update_opt_state(new_pop, new_vals)

        self.changed = True

        if self.opt_algorithm.cur_iter > self.opt_algorithm.itermax:
            # maximum number of iterations exceeded
            # XXX: what return code is appropriate here?
            self.execution.exitcode = os.EX_TEMPFAIL
            return gc3libs.Run.State.TERMINATED

        # still within allowed number of iterations, check convergence
        if self.opt_algorithm.has_converged():
            # report success of sequential task
            self.execution.returncode = 0
            return gc3libs.Run.State.TERMINATED
        else:
            # prepare next evaluation
            self.new_pop = self.opt_algorithm.evolve()
            self.add(
                ComputeTargetVals(
                    self.new_pop,
                    self.jobname,
                    self.opt_algorithm.cur_iter,
                    self.path_to_stage_dir,
                    self.cur_pop_file,
                    self.task_constructor))
            return gc3libs.Run.State.RUNNING

    def __str__(self):
        return self.jobname

    # Adjustments for pickling
    # def __getstate__(self):
    #     state = Task.__getstate__(self)
    # Check that there are no functions in state.
    # for attr in ['opt_algorithm']:
    # 'task_constructor', 'extract_value_fn', 'tasks',
    # del state[attr]
    # state = None
    #     return state

    # def __setstate__(self, state):
    # restore _grid, etc.
    #     Task.__setstate__(self, state)
    # restore loggers
    # self._setup_logging()


class ComputeTargetVals(ParallelTaskCollection):

    """
    :class:`gc3libs.workflow.ParallelTaskCollection` to evaluate the current
      `pop` using the user-supplied :func:`task_constructor`.

    :param pop: Population to evaluate. Must be a NumPy "array-like" value.

    :param str jobname: Name of :class:`GridDriver` instance driving the
      optimization.

    :param int iteration: Current iteration number.

    :param str path_to_stage_dir: Path to directory in which optimization takes
      place.

    :param str cur_pop_file: Filename under which the population is stored in
      the current iteration dir. The population is discarded if no file is
      specified.

    :param task_constructor: Takes a list of x vectors and the path to the
      current iteration directory. Returns Application instances that can be
      executed on the grid.
    """

    def __str__(self):
        return self.jobname

    def __init__(self, pop, jobname, iteration, path_to_stage_dir,
                 cur_pop_file, task_constructor, **extra_args):

        gc3libs.log.debug('entering ComputeTargetVals.__init__')

        # Set up initial variables and set the correct methods.
        self.jobname = jobname + '-' + \
            'compute_target_vals' + '-' + str(iteration)
        self.iteration = iteration

        self.path_to_stage_dir = path_to_stage_dir
        # ComputeTargetVals produces no output.
        # But attribute needs to be specified.
        self.output_dir = path_to_stage_dir
        self.cur_pop_file = cur_pop_file
        self.verbosity = 'DEBUG'
        self.extra_args = extra_args

        # Log activity
        cDate = datetime.date.today()
        cTime = datetime.datetime.time(datetime.datetime.now())
        date_string = '%04d--%02d--%02d--%02d--%02d--%02d' % (
            cDate.year, cDate.month, cDate.day, cTime.hour, cTime.minute, cTime.second)
        gc3libs.log.debug('Establishing parallel task on %s', date_string)

        # Enter an iteration specific folder
        self.iteration_folder = os.path.join(
            self.path_to_stage_dir, 'Iteration-' + str(self.iteration))
        try:
            os.mkdir(self.iteration_folder)
        except OSError:
            print('%s already exists' % self.iteration_folder)

        # save pop to file
        if cur_pop_file:
            np.savetxt(os.path.join(self.iteration_folder, cur_pop_file),
                       pop, delimiter=' ')

        self.tasks = [
            task_constructor(pop_mem, self.iteration_folder) for pop_mem in pop
        ]
        ParallelTaskCollection.__init__(self, self.tasks, **extra_args)



#! /usr/bin/env python
#
"""
Run the BEAST or BEAST-2 programs from a prepared tree of input files.
"""
# Copyright (C) 2012-2016  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'


import os


import gc3libs
from gc3libs.cmdline import SessionBasedScript
from gc3libs.quantity import MiB, GB


def find_files(dirname, suffix=None):
    """
    Return a list of (full paths to) files in the given directory.

    If `suffix` is given, restrict to files with that ending.
    """
    result = []
    for entry in os.listdir(dirname):
        path = os.path.join(dirname, entry)
        if os.path.isfile(path):
            if suffix and not entry.endswith(suffix):
                continue
            result.append(path)
    return result


class GBeastApp(gc3libs.Application):
    def __init__(self, jarfile, dirname, ncores, **extra):
        extra['output_dir'] = os.path.basename(dirname)
        extra['requested_cores'] = ncores
        extra['requested_memory'] = ncores*4*GB

        inputs = find_files(dirname)

        xmls = [path for path in inputs if path.endswith('.xml')]

        treefiles = [path for path in inputs if path.endswith('.trees')]
        if treefiles:
            # Natasha's config places seed in .log/.trees file name
            treename, _ = treefiles[0].split('.')
            parts = treename.split('_')
            self.seed = int(parts[-1])
            self.resume = True
        else:
            self.seed = None
            self.resume = False

        # build command-line
        args = ['java',
                '-Xmx{mem_mb:d}m'.format(mem_mb=extra['requested_memory'].amount(MiB)),
                '-jar', jarfile,
                '-threads', ncores,
                '-beagle_instances', ncores]
        if self.resume:
            args += ['-resume']
        if self.seed:
            args += ['-seed', self.seed]
            args += [xml]

        gc3libs.Application.__init__(
            self,
            arguments = args,
            inputs = [xmls[0]],
            outputs = gc3libs.ANY_OUTPUT,
            stdout = 'stdout.txt',
            stderr = 'stderr.txt',
            **extra)

        
class GBeastScript(SessionBasedScript):
    """
    Script to parallelize execution of BEAST over multiple input files.
    Takes a list of directories containing XML files and processes them all.
    If a `.trees` file is already present in the directory, BEAST is passed
    the ``-resume`` option to append to existing files instead of starting over.
    """
    version = '1.1'
    def setup_options(self):
        self.add_param('-b', '--beast', choices=['beast1', 'beast2'],
                       help='Beast version to run')
        self.add_param('--beast1', default='/apps/BEASTv1.8.2/lib/beast.jar', help='Path to BEAST v1 jar file')
        self.add_param('--beast2', default='/apps/BEASTv2.3.2/lib/beast.jar', help='Path to BEAST v2 jar file')
        self.add_param('path', nargs='+', help='Directory containing XML input files for beast')
        self.add_param('--cores', default=1, type=int, help="Amount of cores to use. Default: %(default)s")

    def parse_args(self):
        if not os.path.isdir(self.params.path):
            raise gc3libs.exceptions.InvalidUsage(
                "{0} is not a directory".format(self.params.path))

    def new_tasks(self, extra):
        tasks = []
        for dirname in self.params.path:
            if self.params.beast:
                beast = self.params.beast
            else:
                xmls = find_files(dirname, '.xml')
                if len(xmls) > 1:
                    self.log.error(
                        "More than 1 `.xml` file in directory `{dirname}`: {xmls!r}"
                        " Skipping it."
                        .format(**locals()))
                    continue
                xml = xmls[0]
                if 'BEAST1' in xml:
                    beast = 'beast1'
                elif 'BEAST2' in xml:
                    beast = 'beast2'
                else:
                    gc3libs.error(
                        "Unable to guess which version of BEAST you want to run."
                        " Skipping file `{xml}`"
                        .format(xml=xml)
                    continue
            jarfile = (self.params.beast1 if beast == 'beast1' else self.params.beast2)
            tasks.append(GBeastApp(jarfile,
                                   dirname,
                                   self.params.cores,
                                   **extra.copy()))
        return tasks

## main

if "__main__" == __name__:
    from gbeast import GBeastScript
    GBeastScript().run()



#! /usr/bin/env python
#
"""
Front-end script for making a session out of a generic command-line.

This is only provided as a quick prototyping tool, and should not be
used for any real and production-level purposes: use the GC3Libs API
and develop a custom tool instead!
"""
# Copyright (C) 2011-2012  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

from __future__ import absolute_import, print_function

__author__ = 'Riccardo Murri <riccardo.murri@gmail.com>'
# summary of user-visible changes
__changelog__ = """
  2012-02-22:
    * Initial draft version.
"""
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gridrun
    gridrun.GridRunScript().run()


## stdlib imports
import itertools
import os
import os.path
import string
import sys

## GC3Libs imports
import gc3libs
from gc3libs import Application, Run, Task
import gc3libs.exceptions
from gc3libs.cmdline import SessionBasedScript, executable_file, nonnegative_int
import gc3libs.utils
from gc3libs.workflow import RetryableTask


## helper class

class SmartApplication(Application):
    """
    Just like the regular `Application` class, except the `arguments`
    list is scanned for strings that name actual existing files or
    directories, which are automatically added to the `inputs` list.
    """
    def __init__(self, arguments, inputs=None, *more_args, **extra_args):
        # convert to string here as we want to compare args to file names
        arguments = [ str(x) for x in arguments ]

        # create `inputs` as would be done in the `Application` class ctor
        if inputs is not None:
            inputs = Application._io_spec_to_dict(
                gc3libs.url.UrlKeyDict, inputs, force_abs=True)
        else:
            inputs = gc3libs.url.UrlKeyDict()

        # scan command-line for things that look like actual files
        executable = arguments[0]
        if os.path.exists(executable):
            executable_name = os.path.basename(executable)
            inputs[executable] = executable_name
            arguments[0] = './' + executable_name
        for i, arg in enumerate(arguments[1:], 1):
            if arg not in inputs and os.path.exists(arg):
                inputs[arg] = os.path.basename(arg)
                arguments[i] = os.path.basename(arg)

        # recurse into superclass ctor
        Application.__init__(self, arguments, inputs, *more_args, **extra_args)


## the script itself

class GridRunScript(SessionBasedScript):
    """
Manage a task session created by parameter substitution into a
generic command-line.

The CMD and ARGs words given in the invocation of this command are
concatenated to form a command-line; in the process, every word that
matches a substitution parameter (defined with the '-D' option, see
below) is given an actual value.  The number of resulting actual
 command lines is the Cartesian product of the sets of all possible
values of substitution parameters.  The whole set of actual
command-lines creates a session; every command in the session is
submitted and managed until successful execution.

Note: substitution parameter names must be UPPERCASE.

If CMD is a path to a local file, then that file will be uploaded
to the remote system as the command to be executed.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            )


    def setup_args(self):
        self.add_param('cmd', metavar='CMD', type=executable_file,
                       help="Path to the command to execute.")
        self.add_param('args', nargs='*', metavar='ARG',
                       help="Arguments for the command to run.")


    def setup_options(self):
        self.add_param("-D", "--define",
                       action="append", dest="define", default=[],
                       help="Define parameters to substitute on the command-line."
                       " Parameter substitution always has the form 'NAME=SPEC',"
                       " where the SPEC part has one of the following three forms."
                       " (1) The string 'from:' followed by file name:"
                       " substitution data will be read from the specified file,"
                       " one value per line."
                       " (2) The string 'list:' followed by a comma-separated list"
                       " of arbitrary values."
                       " (3) The string 'range:' followed by a range specification"
                       " of the form 'LOW:HIGH:STEP': LOW and HIGH are the minimum"
                       " and the maximum values, and STEP is the increment."
                       " The ':STEP' part can be omitted if STEP is 1.")
        self.add_param("-K", "--retry",
                       nargs='?',    # takes one optional argument
                       type=nonnegative_int, metavar='MAX',
                       const=3,      # returned if '-K' is there, but with no argument
                       default=None, # returned if '-K' is *not* there
                       action="store", dest="retry",
                       help="Retry failed jobs up to MAX times."
                       " If MAX is 0, then retry until the jobs succeeds."
                       " By default, failed jobs are *not* retried.")


    def parse_args(self):
        if self.params.retry is not None:
            self.stats_only_for = RetryableTask

        self.subst = { }
        for define in self.params.define:
            try:
                name, spec = define.split('=', 1)
                kind, value = spec.split(':', 1)
            except ValueError:
                raise gc3libs.exceptions.InvalidUsage(
                    "Invalid argument '%s' after -D/--define option."
                    " Parameter substitutions must have the form 'NAME=KIND:VALUE'."
                    % (define,))
            kind = kind.lower()
            # (1) file name
            if 'from' == kind:
                gc3libs.utils.check_file_access(value, os.R_OK,
                                        exception=gc3libs.exceptions.InvalidUsage)
                self.subst[name] = [i.strip() for i in open(value, 'r')]
                gc3libs.log.info(
                    "Reading values for %s from file '%s'", name, value)
            # (2) list of values
            elif 'list' == kind:
                self.subst[name] = value.split(',')
                gc3libs.log.info(
                    "Parameter %s takes values: %s", name, self.subst[name])
            # (3) range
            elif 'range' == kind:
                if value.count(':') == 2:
                    low, high, step = value.split(':')
                elif value.count(':') == 1:
                    low, high = value.split(':')
                    step = '1' # parsed to int or float later on
                else:
                    raise gc3libs.exceptions.InvalidUsage(
                        "Invalid argument '%s' after -D/--define option."
                        " Parameter range substitutions must have the form"
                        " 'NAME=range:LOW:HIGH:STEP'."
                        % (define,))
                # are low, high, step to floats or ints?
                if ('.' in low) or ('.' in high) or ('.' in step):
                    low = float(low)
                    high = float(high)
                    step = float(step)
                else:
                    low = int(low)
                    high = int(high)
                    step = int(step)
                self.subst[name] = gc3libs.utils.irange(low, high, step)
                gc3libs.log.info(
                    "Parameter %s ranges from %s (incl.) to %s (excl.) in increments of %s",
                    name, low, high, step)


    def new_tasks(self, extra):

        # use the name of the executable as job name
        basename = os.path.basename(self.params.cmd)

        # fix an ordering of the subst parameters, independent of any runtime variable
        names = sorted(self.subst.iterkeys())

        inputs = { }

        # decide whether CMD indicates a local file or a command
        # to be searched on the remote systems' PATH
        if os.path.exists(self.params.cmd):
            self.log.info("Uploading local file '%s' as executable.", self.params.cmd)
            gc3libs.utils.check_file_access(self.params.cmd, os.R_OK|os.X_OK)
            executable = './' + os.path.basename(self.params.cmd)
            inputs[os.path.abspath(self.params.cmd)] = os.path.basename(self.params.cmd)
        else:
            if not os.path.isabs(self.params.cmd):
                raise RuntimeError(
                    "You cannot execute a command by calling a relative path,"
                    " because the remote execution directory is empty"
                    " except for files we upload there; but there is"
                    " no file named '%s' here, so I don't know what to upload.",
                    self.params.cmd)
            executable = self.params.cmd

        # create a set of input files for each combination of the
        # substitution parameters
        for values in itertools.product(* (self.subst[name] for name in names)):
            subst = dict((name, value) for name,value in zip(names, values))
            jobname = str.join('_', [basename] + [
                ("%s=%s" % (name, value.translate(None,
                                                  r'\/&|=%$#!?<>()`"' + r"'" + '\a\b\n\r')))
                for name,value in subst.iteritems() ])

            # construct argument list, substituting defined parameters
            arguments = [ executable ]
            for arg in self.params.args:
                if arg in subst:
                    arguments.append(subst[arg])
                else:
                    arguments.append(arg)

            extra_args = extra.copy()
            extra_args['outputs'] = gc3libs.ANY_OUTPUT
            extra_args['output_dir'] = self.make_directory_path(self.params.output, jobname)
            extra_args['stdout'] = jobname + '.stdout.txt'
            extra_args['stderr'] = jobname + '.stderr.txt'
            if self.params.retry is not None:
                yield RetryableTask(
                    SmartApplication(arguments, inputs, **extra_args),
                    self.params.retry,
                    jobname=jobname,
                    **extra_args)
            else:
                yield SmartApplication(arguments, inputs,
                                       jobname=jobname, **extra_args)



#! /usr/bin/env python
#
#   gcrashdetect.py -- Front-end script for evaluating Matlab functions
#   function over a large number of parameters.
#
#   Copyright (C) 2017, 2018  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple Matlab jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gcrashdetect.py --help`` for program usage
instructions.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2017-06-22:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'
__version__ = '1.0'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gcrashdetect
    gcrashdetect.GcrashdetectScript().run()

import os
import sys
import time
import tempfile

import tarfile
import shutil
import pandas

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file, existing_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

DEFAULT_ITERATIONS=1000
DEFAULT_CHUNKS=1
DEFAULT_REMOTE_OUTPUT_FOLDER="results"
MATLAB_CMD="matlab -nosplash -nodisplay -nodesktop -r \'{main_function}({iterations} {mask} {fn} {output});quit\'"
GCRASHDETECT_VALID_INPUT_FILE_EXTENSIONS = ['.m','.cpp','.mat']
GCRASHDETECT_INPUT_ARCHIVE = "gcrashdetect.tgz"

# utility methods
def _scan_and_tar(tarfile_folder, input_folder):
    """
    Create a tar file from all GCRASHDETECT_VALID_INPUT_FILE_EXTENSIONS
    found in input_folder
    """
    try:
        gc3libs.log.debug("Compressing input folder '%s'" % input_folder)
        cwd = os.getcwd()
        os.chdir(input_folder)

        tar = tarfile.open(os.path.join(tarfile_folder,
                                        GCRASHDETECT_INPUT_ARCHIVE),
                           "w:gz",
                           dereference=True)

        for f in [ elem for elem in os.listdir('.') if os.path.splitext(elem)[-1] in GCRASHDETECT_VALID_INPUT_FILE_EXTENSIONS or os.path.isdir(elem)]:
            tar.add(f)
        tar.close()
        os.chdir(cwd)
        return os.path.join(tarfile_folder,GCRASHDETECT_INPUT_ARCHIVE)
    except Exception as x:
        gc3libs.log.error("Failed creating input archive '%s': %s: %s",
                          os.path.join(tarfile_folder, GCRASHDETECT_INPUT_ARCHIVE),
                          x.__class__,x.message)
        raise

## custom application class
class GcrashdetectApplication(Application):
    """
    Execute Matlab main function passing the arguments extracted
    from the main input .csv file.
    """
    application_name = 'gcrashdetect'

    def __init__(self, matlab_function, parameter_list, tarfile=None, **extra_args):

        inputs = dict()
        outputs = dict()

        # execution wrapper needs to be added anyway
        wrapper = resource_filename(Requirement.parse("gc3pie"),
                                    "gc3libs/etc/gcrashdetect_wrapper.sh")

        inputs[wrapper] = "./wrapper.sh"

        arguments = "./wrapper.sh %s " % matlab_function

        for param in parameter_list:
            arguments += " %s " % param

        if tarfile:
            inputs[tarfile] = os.path.basename(tarfile)
            arguments += "-s %s " % inputs[tarfile]

        arguments += DEFAULT_REMOTE_OUTPUT_FOLDER

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = [DEFAULT_REMOTE_OUTPUT_FOLDER],
            stdout = 'gcrashdetect.log',
            join=True,
            executables = "./wrapper.sh",
            **extra_args)


class GcrashdetectScript(SessionBasedScript):
    """
    Parse the input .csv file. For each line in the input .csv file
    create a different Matlab execution.

    The ``gcrashdetect`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gcrashdetect`` will delay submission of
    newly-created jobs so that this limit is never exceeded.

    Once the processing of all chunked files has been completed, ``gcrashdetect``
    aggregates them into a single larger output file located in
    'self.params.output'.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            application = GcrashdetectApplication,
            stats_only_for = GcrashdetectApplication,
            )

    def setup_options(self):
        self.add_param("-d", "--mfuncs", metavar="PATH", type=str,
                       dest="matlab_source_folder", default=None,
                       help="Location of the Matlab scripts and "
                       "related Matlab functions. Default: %(default)s.")

    def setup_args(self):

        self.add_param('matlab_function', type=str,
                       help="Matlab function name")

        self.add_param('csv_input_file', type=existing_file,
                       help="Input .csv file with all parameters to be passed"
                       " to the Matlab function.")

    def parse_args(self):
        """
        Check if in source_folder there is a Matlab file named after
        the Matlab function passed as input argument.
        """
        assert os.path.isfile(os.path.join(self.params.matlab_source_folder,
                                           self.params.matlab_function+'.m')), \
                                           "Matlab function file '%s' not found." % self.params.matlab_function

    def new_tasks(self, extra):
        """
        For each line of the input .csv file generate
        an execution Task
        """
        tasks = []

        tarfile = None
        if self.params.matlab_source_folder:
            tarfile = _scan_and_tar(self.session.path, self.params.matlab_source_folder)


        for parameter in self._enumerate_csv(self.params.csv_input_file):
            parameter_str = '.'.join(str(x) for x in parameter)
            jobname = "gcrashdetect-%s" % parameter_str

            extra_args = extra.copy()

            extra_args['jobname'] = jobname

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', jobname)

            extra_args['matlab_source_folder'] = self.params.matlab_source_folder
            self.log.debug("Creating Application for parameter : %s" %
                           (parameter_str))

            tasks.append(GcrashdetectApplication(
                self.params.matlab_function,
                parameter,
                tarfile,
                **extra_args))

        return tasks

    def _enumerate_csv(self, input_csv):
        """
        For each line of the input .csv file
        return list of parameters
        """
        parameters = pandas.read_csv(input_csv)
        for i,p in enumerate(parameters.values):
            yield p.tolist()



#!/usr/bin/env python
"""
Setup file for installing GC3Pie.
"""

from __future__ import absolute_import
import sys

# ensure we run a "recent enough" version of `setuptools` (CentOS7 still ships
# with setuptools 0.9.8!). There has been some instability in the support for
# PEP-496 environment markers in recent versions of `setuptools`, but
# Setuptools 20.10.0 seems to have restored full support for them, including
# `python_implementation`
from ez_setup import use_setuptools
use_setuptools(version='21.0.0')

import setuptools
import setuptools.dist
# avoid setuptools including `.svn` directories into the PyPI package
from setuptools.command import sdist
try:
    del sdist.finders[:]
except AttributeError:
    # `.finders` was removed from setuptools long ago
    pass


## auxiliary functions
#
def read_whole_file(path):
    """
    Return file contents as a string.
    """
    with open(path, 'r') as stream:
        return stream.read()


## test runner setup
#
# See http://tox.readthedocs.org/en/latest/example/basic.html#integration-with-setuptools-distribute-test-commands # noqa
# on how to run tox when python setup.py test is run
#
from setuptools.command.test import test as TestCommand

class Tox(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import tox
        errno = tox.cmdline(self.test_args)
        sys.exit(errno)


## conditional dependencies
#
# Although PEP-508 and a number of predecessors specify a syntax for
# conditional dependencies in Python packages, support for it is inconsistent
# (at best) among the PyPA tools. An attempt to use the conditional syntax has
# already caused issues #308, #249, #227, and many more headaches to me while
# trying to find a combination of `pip`, `setuptools`, `wheel`, and dependency
# specification syntax that would work reliably across all supported Linux
# distributions. I give up, and revert to computing the dependencies via
# explicit Python code in `setup.py`; this will possibly break wheels but it's
# the least damage I can do ATM.

python_version = sys.version_info[:2]
if python_version == (2, 6):
    version_dependent_requires = [
        # Alternate dependencies for Python 2.6:
        # - PyCLI requires argparse,
        'argparse',
        'lockfile==0.11.0',
        # Paramiko and its dependencies all ceased support for Py2.6
        # NOTE: order matters!  Dependencies must come before the main pkg
        'pycparser<2.19',
        'pynacl<1.1.0',
        'idna<2.7',
        'paramiko<2.4',
        # parsedatetime officially dropped supprt for Py 2.6 in version 1.0
        # but the PyPI tags show that it is compatible with Py2.6 until <=1.4
        'parsedatetime<1.5',
        # python-daemon seems to have dropped Py2.6 between v1.6 and v2.0
        'python-daemon<2.0',
        'pyyaml<=3.11',
        # SQLAlchemy ceased support for Py 2.6 in version 1.2.0
        'sqlalchemy<1.2',
    ]
else:
    version_dependent_requires = [
        'lockfile',
        'paramiko', 'pycrypto',
        # Needed for parsing human-readable dates (gselect uses it).
        'parsedatetime',
        # Needed by `gc3libs.cmdline`
        'python-daemon',
        'pyyaml',
        'sqlalchemy',
    ]


## real setup description begins here
#
setuptools.setup(
    name="gc3pie",
    version="2.6.5",  # see PEP 440

    packages=setuptools.find_packages(exclude=['ez_setup']),
    # metadata for upload to PyPI
    description=(
        "A Python library and simple command-line frontend for"
        " computational job submission to multiple resources."
    ),
    long_description=read_whole_file('README.rst'),
    author=', '.join([
        # only long-time core authors are listed here;
        # please see file `docs/credits.rst`
        'Sergio Maffioletti',
        'Antonio Messina',
        'Riccardo Murri',
    ]),
    author_email="gc3pie-dev@googlegroups.com",
    license="LGPL",
    keywords=str.join(' ', [
        "batch",
        "cloud",
        "cluster",
        "differential optimization",
        "ec2",
        "gridengine",
        "job management",
        "large-scale data analysis",
        "openstack",
        "pbs",
        "remote execution",
        "sge",
        "slurm",
        "ssh",
        "torque",
        "workflow",
    ]),
    url="http://gc3pie.readthedocs.io/",  # project home page

    # see http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        ("License :: OSI Approved :: GNU Library or"
         " Lesser General Public License (LGPL)"),
        "License :: DFSG approved",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: System :: Distributed Computing",
        ],

    entry_points={
        'console_scripts': [
            # the generic, catch-all script:
            'gc3utils = gc3utils.frontend:main',
            # entry points to specific subcommands:
            'gclean = gc3utils.frontend:main',
            'gclient = gc3utils.frontend:main',
            'gget = gc3utils.frontend:main',
            'ginfo = gc3utils.frontend:main',
            'gkill = gc3utils.frontend:main',
            'gservers = gc3utils.frontend:main',
            'gresub = gc3utils.frontend:main',
            'gstat = gc3utils.frontend:main',
            'gsub = gc3utils.frontend:main',
            'gtail = gc3utils.frontend:main',
            'gsession = gc3utils.frontend:main',
            'gselect = gc3utils.frontend:main',
            'gcloud = gc3utils.frontend:main',
            ],
    },

    # run-time dependencies
    install_requires=(version_dependent_requires + [
        'blinker',
        'coloredlogs',
        'dictproxyhack',
        # prettytable -- format tabular text output
        'prettytable',
        # pyCLI -- object-oriented command-line app programming
        'pyCLI',
        # needed by DependentTaskCollection
        # (but incompatible with Py 2.6, so we include a patched copy)
        #toposort==1.0
        # Py2/Py3 compatibility layer
        'future>=0.17.0',
        'six',  # only used in `gc3libs/quantity.py`
    ]),
    extras_require={
        'ec2': [
            # The following Python modules are required by GC3Pie's `ec2`
            # resource backend.
            'boto',
        ],
        'daemon': [
            # daemon and inotifyx required for SessionBasedDaemon but
            # inotify functionality is only available on Linux, so
            # this is an optional feature ...
            'inotify_simple',
        ],
        'openstack': [
            'python-keystoneclient',
            'python-glanceclient',
            'python-neutronclient',
            'python-novaclient',
            'python-swiftclient',
            'os-client-config',
        ],
        'optimizer': [
            # The following Python modules are required by GC3Pie's
            # `gc3libs.optimizer` module.
            'numpy',
        ],
    },

    # Apparently, this list is read from bottom to top...
    tests_require=[
        'tox',
        'pytest-catchlog',
        'pytest>=4.1',
        'mock',
    ],
    cmdclass={'test': Tox},

    # additional non-Python files to be bundled in the package
    package_data={
        'gc3libs': [
            # helper scripts for classes in the core library
            'etc/codeml.pl',
            'etc/rosetta.sh',
            'etc/run_R.sh',
            'etc/turbomole.sh',
            'etc/run_matlab.sh',
            # Downloader script
            'etc/downloader.py',
            # example files
            'etc/gc3pie.conf.example',
            'etc/logging.conf.example',
            'etc/minibacct.c',
            # helper scripts for specific gc3apps
            'etc/gbenchmark_infomap_wrapper.sh',
            'etc/gbenchmark_wrapper.sh',
            'etc/gbenchmark_wrapper_allinone.py',
            'etc/gcelljunction_wrapper.sh',
            'etc/geosphere_wrapper.sh',
            'etc/geotop_wrap.sh',
            'etc/geotop_wrap.sh',
            'etc/gmodis_wrapper.sh',
            'etc/gndn_wrapper.sh',
            'etc/gnfs-cmd',
            'etc/gnlp_wrapper.py',
            'etc/gpyrad_wrapper.sh',
            'etc/grdock_wrapper.sh',
            'etc/gstructure_wrapper.sh',
            'etc/gweight_wrap.sh',
            'etc/gwrappermc_wrapper.sh',
            'etc/run_gtsub_control.sh',
            'etc/smd_projections_wrapper.sh',
            'etc/square.sh',
        ],
    },
    data_files=[
        ('gc3apps', [
            'gc3apps/gc3.uzh.ch/gridrun.py',
            'gc3apps/codeml/gcodeml.py',
            'gc3apps/gamess/grundb.py',
            'gc3apps/gamess/ggamess.py',
            ]),
    ],

    # `zip_safe` can ease deployment, but is only allowed if the package
    # do *not* do any __file__/__path__ magic nor do they access package data
    # files by file name (use `pkg_resources` instead).
    zip_safe=True,
)



#! /usr/bin/env python

"""
Job control on SGE clusters (possibly connecting to the front-end via SSH).
"""

# Copyright (C) 2009-2014, 2016  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
from builtins import str
__docformat__ = 'reStructuredText'


# stdlib imports
import datetime
import math
import re

from collections import defaultdict

# GC3Pie imports
from gc3libs import log, Run
from gc3libs.backends import LRMS
import gc3libs.exceptions
from gc3libs.quantity import Memory, kB, MB, GB
from gc3libs.quantity import seconds
import gc3libs.utils
from gc3libs.utils import (same_docstring_as, sh_quote_safe_cmdline,
                           sh_quote_unsafe_cmdline)
from . import batch


# auxiliary functions

def _int_floor(val):
    """Return `val` rounded to nearest integer towards 0."""
    return int(math.floor(float(val)))


def _to_duration(val):
    """Convert a floating point number of seconds to a
    `gc3libs.quantity.Duration` value."""
    try:
        return float(val) * seconds
    except Exception as err:
        gc3libs.log.warning(
            "Grid Engine backend:"
            " Cannot interpret '%s' as a duration (time unit: seconds):"
            " %s: %s.",
            err.__class__.__name__, str(err), val)
        return None


def _to_memory(val):
    """Convert a Grid Engine MEMORY value to a
    `gc3libs.quantity.Memory` one."""
    try:
        unit = val[-1]
        if unit in ['G', 'g']:
            return float(val[:-2]) * GB
        elif unit in ['M', 'm']:
            return float(val[:-2]) * MB
        elif unit in ['K', 'k']:
            return float(val[:-2]) * kB
        else:
            # SGE's default is bytes
            return float(val) * Memory.B
    except Exception:
        gc3libs.log.warning("Grid Engine backend: Cannot interpret '%s' "
                            "as a MEMORY value.", val)
        return None


# `_convert` is a `dict` instance, mapping key names to functions
# that parse a value from a string into a Python native type.
_convert = {
    'slots': int,
    'slots_used': int,
    'slots_resv': int,
    'slots_total': int,
    'load_avg': float,
    'load_short': float,
    'load_medium': float,
    'load_long': float,
    'np_load_avg': float,
    'np_load_short': float,
    'np_load_medium': float,
    'np_load_long': float,
    'num_proc': _int_floor,  # SGE considers `num_proc` a
    # floating-point value...
    'swap_free': gc3libs.utils.to_bytes,
    'swap_total': gc3libs.utils.to_bytes,
    'swap_used': gc3libs.utils.to_bytes,
    'mem_free': gc3libs.utils.to_bytes,
    'mem_used': gc3libs.utils.to_bytes,
    'mem_total': gc3libs.utils.to_bytes,
    'virtual_free': gc3libs.utils.to_bytes,
    'virtual_used': gc3libs.utils.to_bytes,
    'virtual_total': gc3libs.utils.to_bytes,
}


def _parse_value(key, value):
    try:
        return _convert[key](value)
    except:
        return value


def _parse_asctime(val):
    """
    Parse a timestamp string output by GE and return a Python `datetime` object.
    """
    # while the format of asctime() is strictly mandated by POSIX and
    # the ISO C standard, Univa GE changed the date/time printing in
    # `qacct -j` to use the format of `sge_ctime()` which hard-code a
    # MM/DD/YYYY representation.  Yet another representation (ISO
    # 8601) is given in XML output... so try them all, one at a time
    # hoping one succeeds!
    for fmt in [
            '%a %b %d %H:%M:%S %Y',  # standard asctime() format
            '%m/%d/%Y %H:%M:%S',     # sge_ctime()
            '%Y-%m-%dT%H:%M:%S',     # ISO 8601 / sge_ctimeXML()
    ]:
        try:
            return datetime.datetime.strptime(val, fmt)
        except ValueError:
            pass  # try next format
    gc3libs.log.error(
        "Cannot parse '%s' as a SGE-format time stamp", val)
    return None


def parse_qstat_f(qstat_output):
    """
    Parse SGE's ``qstat -F`` output (as contained in string `qstat_output`)
    and return a `dict` instance, mapping each queue name to its attributes.
    """
    # a job report line starts with a numeric job ID
    # _job_line_re = re.compile(r'^[0-9]+ \s+', re.X)
    # queue report header line starts with queuename@hostname
    _queue_header_re = re.compile(
        r'^([a-z0-9\._-]+)@([a-z0-9\.-]+) \s+ ([BIPCTN]+) '
        '\s+ ([0-9]+)?/?([0-9]+)/([0-9]+)',
        re.I | re.X)
    # property lines always have the form 'xx:propname=value'
    _property_line_re = re.compile(r'^[a-z]{2}:([a-z_]+)=(.+)', re.I | re.X)

    def dzdict():
        def zdict():
            return defaultdict(int)
        return defaultdict(zdict)
    result = defaultdict(dzdict)
    qname = None
    for line in qstat_output.split('\n'):
        # strip leading and trailing whitespace
        line = line.strip()
        # is this a queue header?
        match = _queue_header_re.match(line)
        if match:
            qname, hostname, kind, slots_resv, slots_used, slots_total \
                = match.groups()
            if 'B' not in kind:
                continue  # ignore non-batch queues

            # Some versions of SGE do not have a "reserved" digit in
            # the slots column, so slots_resv will be set to None.
            # For our purposes it is better that it is 0.
            if slots_resv is None:
                slots_resv = 0
            # key names are taken from 'qstat -xml' output
            result[qname][hostname]['slots_resv'] = _parse_value(
                'slots_resv', slots_resv)
            result[qname][hostname]['slots_used'] = _parse_value(
                'slots_used', slots_used)
            result[qname][hostname]['slots_total'] = _parse_value(
                'slots_total', slots_total)
        # is this a property line?
        match = _property_line_re.match(line)
        if match:
            key, value = match.groups()
            result[qname][hostname][key] = _parse_value(key, value)
    return result


def compute_nr_of_slots(qstat_output):
    """
    Compute the number of total, free, and used/reserved slots from
    the output of SGE's ``qstat -F``.

    Return a dictionary instance, mapping each host name into a
    dictionary instance, mapping the strings ``total``, ``available``,
    and ``unavailable`` to (respectively) the the total number of
    slots on the host, the number of free slots on the host, and the
    number of used+reserved slots on the host.

    Cluster-wide totals are associated with key ``global``.

    **Note:** The 'available slots' computation carried out by this
    function is unreliable: there is indeed no notion of a 'global' or
    even 'per-host' number of 'free' slots in SGE.  Slot numbers can
    be computed per-queue, but a host can belong in different queues
    at the same time; therefore the number of 'free' slots available
    to a job actually depends on the queue it is submitted to.  Since
    SGE does not force users to submit explicitly to a queue, rather
    encourages use of a sort of 'implicit' routing queue, there is no
    way to compute the number of free slots, as this entirely depends
    on how local policies will map a job to the available queues.
    """
    qstat = parse_qstat_f(qstat_output)

    def zero_initializer():
        return 0

    def dict_with_zero_initializer():
        return defaultdict(zero_initializer)
    result = defaultdict(dict_with_zero_initializer)
    for q in qstat.keys():
        for host in qstat[q].keys():
            r = result[host]
            s = qstat[q][host]
            r['total'] = max(s['slots_total'], r['total'])
            r['unavailable'] = max(s['slots_used'] + s['slots_resv'],
                                   r['unavailable'])
    # compute available slots by subtracting the number of
    # used+reserved from the total
    g = result['global']
    for host in result.keys():
        r = result[host]
        r['available'] = r['total'] - r['unavailable']
        # update cluster-wide ('global') totals
        g['total'] += r['total']
        g['unavailable'] += r['unavailable']
        g['available'] += r['available']
    return result


def parse_qhost_f(qhost_output):
    """
    Parse SGE's ``qhost -F`` output (as contained in string `qhost_output`)
    and return a `dict` instance, mapping each host name to its attributes.
    """
    result = defaultdict(dict)
    n = 0
    for line in qhost_output.split('\n'):
        # skip header lines
        n += 1
        if n < 3:
            continue
        if not line.startswith(' '):
            # host lines begin at column 0
            hostname = line.split(' ')[0]
        else:
            # property lines start with TAB
            key, value = line.split('=')
            ignored, key = key.split(':')
            result[hostname][key] = _parse_value(key, value)
    return result


def count_jobs(qstat_output, whoami):
    """
    Parse SGE's ``qstat`` output (as contained in string `qstat_output`)
    and return a quadruple `(R, Q, r, q)` where:

      * `R` is the total number of running jobs in the SGE cell (from
        any user);

      * `Q` is the total number of queued jobs in the SGE cell (from
        any user);

      * `r` is the number of running jobs submitted by user `whoami`;

      * `q` is the number of queued jobs submitted by user `whoami`
    """
    total_running = 0
    total_queued = 0
    own_running = 0
    own_queued = 0
    n = 0
    for line in qstat_output.split('\n'):
        # skip header lines
        n += 1
        if n < 3:
            continue
        # remove leading and trailing whitespace
        line = line.strip()
        if len(line) == 0:
            continue
        jid, prio, name, user, state, rest = re.split(r'\s+', line, 5)
        # skip in error/hold/suspended/deleted state
        if (('E' in state) or ('h' in state) or ('T' in state)
                or ('s' in state) or ('S' in state) or ('d' in state)):
            continue
        if 'q' in state:
            total_queued += 1
            if user == whoami:
                own_queued += 1
        if 'r' in state:
            total_running += 1
            if user == whoami:
                own_running += 1
    return (total_running, total_queued, own_running, own_queued)


# FIXME: I think this function is completely wrong and only exists to
# support GAMESS' ``qgms``, which does not allow users to specify the
# name of STDOUT/STDERR files.  When we have a standard flexible
# submission mechanism for all applications, we should remove it!
def _sge_filename_mapping(jobname, jobid, file_name):
    """
    Map STDOUT/STDERR filenames (as recorded in `Application.outputs`)
    to SGE/OGS default STDOUT/STDERR file names (e.g.,
    ``<jobname>.o<jobid>``).
    """
    try:
        return {
            # XXX: SGE-specific?
            ('%s.out' % jobname): ('%s.o%s' % (jobname, jobid)),
            ('%s.err' % jobname): ('%s.e%s' % (jobname, jobid)),
            # FIXME: the following is definitely GAMESS-specific
            ('%s.cosmo' % jobname): ('%s.o%s.cosmo' % (jobname, jobid)),
            ('%s.dat' % jobname): ('%s.o%s.dat' % (jobname, jobid)),
            ('%s.inp' % jobname): ('%s.o%s.inp' % (jobname, jobid)),
            ('%s.irc' % jobname): ('%s.o%s.irc' % (jobname, jobid))}[file_name]
    except KeyError:
        return file_name


_qsub_jobid_re = re.compile(r'Your job (?P<jobid>\d+) '
                            '\("(?P<jobname>.+)"\) has been submitted', re.I)
"""
Regex for extracting the job number and name from Grid Engine's `qsub` output.
"""


class SgeLrms(batch.BatchSystem):

    """
    Job control on SGE clusters (possibly by connecting via SSH to a
    submit node).
    """

    _batchsys_name = 'Grid Engine'

    def __init__(self, name,
                 # this are inherited from the base LRMS class
                 architecture, max_cores, max_cores_per_job,
                 max_memory_per_core, max_walltime,
                 auth,  # ignored if `transport` is 'local'
                 # these are inherited from the `BatchSystem` class
                 frontend, transport,
                 # these are specific to the SGE class
                 default_pe=None,
                 # (Note that optional arguments to the `BatchSystem` class,
                 # e.g.:
                 #     keyfile=None, accounting_delay=15,
                 # are collected into `extra_args` and should not be explicitly
                 # spelled out in this signature.)
                 **extra_args):

        # init base class
        batch.BatchSystem.__init__(
            self, name,
            architecture, max_cores, max_cores_per_job,
            max_memory_per_core, max_walltime, auth,
            frontend, transport, **extra_args)

        self.default_pe = default_pe

        self.qsub = self._get_command_argv('qsub')

        # GridEngine commands
        self._qacct = self._get_command('qacct')
        self._qdel = self._get_command('qdel')
        self._qstat = self._get_command('qstat')

    def _submit_command(self, app):
        sub_argv, app_argv = app.qsub_sge(self)
        return (sh_quote_safe_cmdline(sub_argv),
                sh_quote_unsafe_cmdline(app_argv))

    def _parse_submit_output(self, output):
        """Parse the ``qsub`` output for the local jobid."""
        return self.get_jobid_from_submit_output(output, _qsub_jobid_re)

    def _stat_command(self, job):
        return ("%s | egrep  '^ *%s'" % (self._qstat, job.lrms_jobid))

    def _parse_stat_output(self, stdout, stderr):
        ge_status_code = stdout.split()[4]
        log.debug(
            "translating SGE's `qstat` code '%s' to gc3libs.Run.State",
            ge_status_code)
        if (ge_status_code in ['s', 'S', 'T']
            or ge_status_code.startswith('h')):
            state = Run.State.STOPPED
        elif 'qw' in ge_status_code:
            state = Run.State.SUBMITTED
        elif ('r' in ge_status_code
              or 'R' in ge_status_code
              or 't' in ge_status_code):
            state = Run.State.RUNNING
        elif ge_status_code == 'E':  # error condition
            state = Run.State.TERMINATING
        else:
            log.warning("unknown SGE job status '%s', returning `UNKNOWN`",
                        ge_status_code)
            state = Run.State.UNKNOWN
        # to get the exit status information we'll have to parse
        # `qacct` output so put ``None`` here
        return self._stat_result(state, None)

    def _acct_command(self, job):
        return ("%s -j %s" % (self._qacct, job.lrms_jobid))

    _qacct_keyval_mapping = {
        # qacct field name
        # |              `Task.execution` attribute
        # |              |
        # |              |                       converter function
        # |              |                       |
        # |              |                       |
        #   ... common backend attrs (see Issue 78) ...
        'slots':         ('cores',               int),
        'exit_status':   ('exitcode',            int),
        'cpu':           ('used_cpu_time',       _to_duration),
        'ru_wallclock':  ('duration',            _to_duration),
        'maxvmem':       ('max_used_memory',     _to_memory),
        #   ... SGE-only attrs ...
        'end_time':      ('sge_completion_time', _parse_asctime),
        'failed':        ('sge_failed',          int),
        'granted_pe':    ('sge_granted_pe',      str),
        'hostname':      ('sge_hostname',        str),
        'jobname':       ('sge_jobname',         str),
        'qname':         ('sge_queue',           str),
        'qsub_time':     ('sge_submission_time', _parse_asctime),
        'start_time':    ('sge_start_time', _parse_asctime),
    }

    def _parse_acct_output(self, stdout, stderr):
        acctinfo = {}
        for line in stdout.split('\n'):
            # skip empty and header lines
            line = line.strip()
            if line == '' or '===' in line:
                continue
            # extract key/value pairs from `qacct` output
            key, value = line.split(' ', 1)
            value = value.strip()
            if key == 'failed':
                # value may be, e.g., "100 : assumedly after job"
                value = value.split()[0]
            try:
                dest, conv = self._qacct_keyval_mapping[key]
                acctinfo[dest] = conv(value)
            except KeyError:
                # no conversion by default -- keep it a string
                acctinfo['sge_' + key] = value
            except (ValueError, TypeError) as err:
                log.error(
                    "Cannot parse value '%s' for qacct parameter '%s': %s: %s",
                    value, key, err.__class__.__name__, str(err))
                acctinfo[dest] = None
        assert 'exitcode' in acctinfo, (
            "Could not extract exit code from `tracejob` output")
        acctinfo['termstatus'] = Run.shellexit_to_returncode(
            acctinfo.pop('exitcode'))
        return acctinfo

    def _cancel_command(self, jobid):
        return ("%s %s" % (self._qdel, jobid))

    @same_docstring_as(LRMS.get_resource_status)
    @LRMS.authenticated
    def get_resource_status(self):
        try:
            self.transport.connect()

            _command = ("%s -U %s" % (self._qstat, self._username))
            log.debug("Running `%s`...", _command)
            exit_code, qstat_stdout, stderr \
                = self.transport.execute_command(_command)
            if exit_code != 0:
                # cannot continue
                raise gc3libs.exceptions.LRMSError(
                    "SGE backend failed executing '%s':"
                    "exit code: %d; stdout: '%s'; stderr: '%s'." %
                    (_command, exit_code, qstat_stdout, stderr))

            _command = ("%s -F -U %s" % (self._qstat, self._username))
            log.debug("Running `%s`...", _command)
            exit_code, qstat_F_stdout, stderr \
                = self.transport.execute_command(_command)
            if exit_code != 0:
                # cannot continue
                raise gc3libs.exceptions.LRMSError(
                    "SGE backend failed executing '%s':"
                    "exit code: %d; stdout: '%s'; stderr: '%s'." %
                    (_command, exit_code, qstat_F_stdout, stderr))

            (total_running, self.queued, self.user_run, self.user_queued) \
                = count_jobs(qstat_stdout, self._username)
            slots = compute_nr_of_slots(qstat_F_stdout)
            self.free_slots = int(slots['global']['available'])
            self.used_quota = -1

            log.info("Updated resource '%s' status:"
                     " free slots: %d,"
                     " own running jobs: %d,"
                     " own queued jobs: %d,"
                     " total queued jobs: %d",
                     self.name,
                     self.free_slots,
                     self.user_run,
                     self.user_queued,
                     self.queued,
                     )
            return self

        except Exception as ex:
            log.error("Error querying remote LRMS, see debug log for details.")
            log.debug("Error querying LRMS: %s: %s",
                      ex.__class__.__name__, str(ex))
            raise

# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="sge",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
#   gsPhenotypicalHomologyExample.py -- Front-end script for running sheepriver simulation
#   written in java.
#
#   Copyright (C) 2016  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gsPhenotypicalHomologyExample.py --help`` for program usage
instructions.

Input argument consists of:
- hunting range: integer range (e.g. 1:430)
...

Options:
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2016-03-29:
  * Initial version
  2016-05-06:
  * TODO: allow 0 as hunting value
  * TODO: check consistency of seed file
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gsPhenotypicalHomologyExample
    gsPhenotypicalHomologyExample.GsPhenotypicalHomologyExampleScript().run()

import os
import sys
import time
import tempfile
import re

import shutil
# import csv

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

DEFAULT_ITERATIONS=10000
DEFAULT_OUTPUT_FOLDER="results"
DEFAULT_OUTPUT_ARCHIVE="results.tgz"

## custom application class
class GsPhenotypicalHomologyExampleApplication(Application):
    """
    """
    application_name = 'gsphenotypicalhomologyexample'

    def __init__(self, hunting, **extra_args):
        """
        Remote execution: ./sheepriver_wrapper.sh 77
                          -p param_SheepRiver_wear
                          -m listOfRandomSeeds
                          -i 10
                          -j src/StartSimulation.jar
        """

        inputs = dict()
        outputs = dict()

        wrapper = resource_filename(Requirement.parse("gc3pie"),
                                    "gc3libs/etc/sheepriver_wrapper.sh")
        inputs[wrapper] = "./wrapper.sh"

        arguments = "./wrapper.sh %d " % hunting

        if 'sources' in extra_args:
            sources = os.path.basename(extra_args['sources'])
            inputs[extra_args['sources']] = sources
            arguments += " -s %s " % sources

        if 'sheepriver' in extra_args:
            sheepriver = os.path.basename(extra_args['sheepriver'])
            inputs[extra_args['sheepriver']] = sheepriver
            arguments += " -p %s " % sheepriver

        if 'seeds' in extra_args:
            seeds = os.path.basename(extra_args['seeds'])
            inputs[extra_args['seeds']] = seeds
            arguments += " -m %s " % seeds

        if 'jar' in extra_args:
            jar = os.path.basename(extra_args['jar'])
            inputs[extra_args['jar']] = jar
            arguments += " -j %s " % jar

        if 'iterations' in extra_args:
            arguments += " -i %d " % extra_args['iterations']

        # set output folder
        arguments += " -o %s " % DEFAULT_OUTPUT_ARCHIVE

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = [DEFAULT_OUTPUT_ARCHIVE],
            stdout = 'gsPhenotypicalHomologyExample.log',
            join=True,
            executables = ["./wrapper.sh"],
            **extra_args)

class GsPhenotypicalHomologyExampleScript(SessionBasedScript):
    """

    The ``gsPhenotypicalHomologyExample`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gsPhenotypicalHomologyExample`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GsPhenotypicalHomologyExampleApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GsPhenotypicalHomologyExampleApplication,
            )

    def setup_args(self):

        self.add_param('range', type=str,
                       help="hunting pressures range. "
                       "Format: [int],[int]|[int]:[int]. E.g 1:432|3|6,9")

    def setup_options(self):
        self.add_param("-S", "--source", metavar="STRING", type=str,
                       dest="sources",
                       default=None,
                       help="Location of java scripts to drive the "
                       " execution of sheepriver. Default: %(default)s")

        self.add_param("-j", "--jar", metavar="STRING", type=str,
                       dest="jar",
                       default=None,
                       help="Location of .jar package to drive the "
                       " execution of sheepriver. Default: %(default)s")

        self.add_param("-P", "--param_SheepRiver_wear", metavar="PATH", type=str,
                       dest="param_SheepRiver_wear",
                       default=None,
                       help="Location of the 'param_SheepRiver_wear' input file. "
                       "Default: %(default)s")

        self.add_param("-M", "--seeds", metavar="PATH", type=str,
                       dest="seeds",
                       default=None,
                       help="Location of the seeds file. "
                       "Default: %(default)s")

        self.add_param("-I", "--replications", metavar="INT", type=int,
                       dest="iterations",
                       default=DEFAULT_ITERATIONS,
                       help="Number of repeating iterations. "
                       "Default: %(default)s")

    def parse_args(self):
        try:
            if self.params.sources:
                assert os.path.isdir(self.params.sources), \
                    "Simulation source folder %s not found" % self.params.sources


            if self.params.param_SheepRiver_wear:
                assert os.path.isfile(self.params.param_SheepRiver_wear), \
                    "SheepRiver file %s not found" % self.params.param_SheepRiver_wear

            if self.params.seeds:
                assert os.path.isfile(self.params.seeds), \
                    "Seeds file %s not found" % self.params.seeds

            if self.params.jar:
                assert os.path.isfile(self.params.jar), \
                    "Jar file %s not found" % self.params.jar


            assert isinstance(int(self.params.iterations),int), \
                "Iterations should be a positive integer"

            # Validate month range
            try:
                # Check whether only single value has been passed
                try:
                    assert isinstance(int(self.params.range),int)
                    self.input_range = [int(self.params.range)]
                except ValueError as ex:
                    # Identify the separator
                    if len(self.params.range.split(":")) == 2:
                        # Use ':' as separator
                        try:
                            start,end = [ int(mrange) for mrange in self.params.range.split(":") ]
                            if end <= start:
                                raise ValueError("No valid input range. "
                                                 "Format: [int],[int]|[int]:[int]. E.g 1:432|3")
                            self.input_range = range(start,end+1)
                        except (TypeError, ValueError) as ex:
                            gc3libs.log.critical(ex.message)
                            raise ValueError("No valid input range. "
                                             "Format: [int],[int]|[int]:[int]. E.g 1:432|3")
                    elif len(self.params.range.split(",")) > 0:
                        # Use ',' as separator
                        self.input_range = [ int(mrange) for mrange in self.params.range.split(",") ]
                    else:
                        # Anything else should fail
                        raise ValueError("No valid input range. "
                                         "Format: [int],[int]|[int]:[int]. E.g 1:432|3")

            except ValueError as ex:
                gc3libs.log.debug(ex.message)
                raise AttributeError("No valid input range. "
                                     "Format: [int],[int]|[int]:[int]. E.g 1:432|3")

        except AssertionError as ex:
            raise OSError(ex.message)

    def new_tasks(self, extra):
        """
        For each input folder, create an instance of GsPhenotypicalHomologyExampleApplication
        """
        tasks = []

        for hunting in self.input_range:

            # filename example: 0103645_anat.nii.gz
            # extract root folder name to be used as jobname
            extra_args = extra.copy()
            extra_args['jobname'] = 'sheepriver-%s' % str(hunting)

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME',
                                                                        'run_%s' % hunting)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION',
                                                                        'run_%s' % hunting)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE',
                                                                        'run_%s' % hunting)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME',
                                                                        'run_%s' % hunting)

            if self.params.sources:
                extra_args['sources'] = os.path.abspath(self.params.sources)

            if self.params.param_SheepRiver_wear:
                extra_args['sheepriver'] = os.path.abspath(self.params.param_SheepRiver_wear)

            if self.params.seeds:
                extra_args['seeds'] = os.path.abspath(self.params.seeds)

            if self.params.jar:
                extra_args['jar'] = os.path.abspath(self.params.jar)

            extra_args['iterations'] = self.params.iterations

            tasks.append(GsPhenotypicalHomologyExampleApplication(
                hunting,
                **extra_args))

        return tasks



#! /usr/bin/env python

"""
Facade to store and retrieve Job information from permanent storage.

A usage warning
~~~~~~~~~~~~~~~

This module saves Python objects using the `pickle` framework: thus,
the `Application` subclass corresponding to a job must be already
loaded (or at least ``import``-able) in the Python interpreter for
`pickle` to be able to 'undump' the object from its on-disk
representation.

In other words, if you create a custom `Application` subclass in some
client code, GC3Utils won't be able to read job files created by this
code, because the class definition is not available in GC3Utils.

The recommended simple workaround is for a stand-alone script to
'import self' and then use the fully qualified name to run the script.
In other words, start your script with this boilerplate code::

    if __name__ == '__main__':
        import myscriptname
        myscriptname.MyScript().run()

The rest of the script now runs as the ``myscript`` module, which does
the trick!

.. note::

  Of course, the ``myscript.py`` file must be in the search path of
  the Python interpreter, or GC3Utils will still complain!

"""

# Copyright (C) 2009-2012, 2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
__docformat__ = 'reStructuredText'


# Export the "public API" towards other modules, so that
# one can do ``import gc3libs.persistence`` and load whatever
# should be normally needed and supported.  Other modules in
# this package should be considered "internal use only".
from .filesystem import FilesystemStore
from .idfactory import IdFactory, JobIdFactory
from .serialization import Persistable
from .store import make_store

__all__ = ['make_store', 'Persistable', 'IdFactory',
           'JobIdFactory', 'FilesystemStore']

# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="persistence",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2012  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
from builtins import range
from builtins import object
__docformat__ = 'reStructuredText'

import logging
import os
import shutil
import tempfile

from gc3libs import Application, Run, configure_logger, create_engine
from gc3libs.workflow import SequentialTaskCollection

loglevel = logging.ERROR
configure_logger(loglevel, "test_issue_335")


class MySequentialCollection(SequentialTaskCollection):

    def __init__(self, *args, **kwargs):
        SequentialTaskCollection.__init__(self, *args, **kwargs)
        self.next_called_n_times = 0

    def next(self, x):
        """count times next() is called"""
        self.next_called_n_times += 1
        return SequentialTaskCollection.next(self, x)


class test_issue_335(object):

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        CONF_FILE = """
[auth/dummy]
type = ssh
username = dummy

[resource/localhost]
enabled = true
auth = dummy
type = shellcmd
frontend = localhost
transport = local
max_cores_per_job = 2
max_memory_per_core = 2
max_walltime = 8
max_cores = 2
architecture = x86_64
override = False
resourcedir = %s
"""
        self.cfgfile = os.path.join(self.tmpdir, 'gc3pie.conf')
        self.resourcedir = os.path.join(self.tmpdir, 'shellcmd.d')
        fp = open(self.cfgfile, 'w')
        fp.write(CONF_FILE % self.resourcedir)
        fp.close()

    def test_issue_335(self):
        """Test that SequentialTasksCollection goes in TERMINATED state when
        all of its tasks are in TERMINATED state."""
        num_tasks_in_seq = 5
        seq = MySequentialCollection([
            Application(
                ['echo', 'test1'],
                [], [],
                os.path.join(self.tmpdir, 'test.%d.d' % i))
            for i in range(num_tasks_in_seq)
        ])
        engine = create_engine(self.cfgfile, auto_enable_auth=True)
        engine.add(seq)
        while True:
            engine.progress()
            if (len([task for task in seq.tasks
                     if task.execution.state == Run.State.TERMINATED])
                    == num_tasks_in_seq):
                engine.progress()
                # check that final SequentialCollection state is TERMINATED
                assert seq.execution.state == Run.State.TERMINATED
                break
        # check that next() has been called once per each task
        assert seq.next_called_n_times == num_tasks_in_seq

    def tearDown(self):
        shutil.rmtree(self.tmpdir)


if "__main__" == __name__:
    import pytest
    pytest.main(["-v", __file__])



#! /usr/bin/env python

r"""
This module implements a global optimization algorithm called Differential
Evolution.

Consider the following optimization problem: :math:`min ~ f(\mathbf{x}) ~~ s.t.
~~ \mathbf{x} \\in D`, where :math:`D \\in \mathbb{R}^d` and :math:`f: D
\mapsto \mathbb{R}`. Class :class:`DifferentialEvolutionAlgorithm
<gc3libs.optimizer.dif_evolution.DifferentialEvolutionAlgorithm>` solves this
optimization problem using the differential evolution algorithm. No further
assumptions on the function :math:`f` are needed. Thus it can be non-convex,
noisy etc.

The domain :math:`D` is implicitly specified by passing the function
:func:`filtern_fn` to :class:`DifferentialEvolutionAlgorithm`.

Some information related to Differential Evolution can be found in the following papers:

1) Tvrdik 2008: http://www.proceedings2008.imcsit.org/pliks/95.pdf
2) Fleetwood: http://www.maths.uq.edu.au/MASCOS/Multi-Agent04/Fleetwood.pdf
3) Piyasatian: http://www-personal.une.edu.au/~jvanderw/DE_1.pdf

`~gc3libs.optimizer.dif_evolution.DifferentialEvolutionAlgorithm.evolve_fn`:func: is an adaptation of the following MATLAB code:
http://www.icsi.berkeley.edu/~storn/DeMat.zip hosted on http://www.icsi.berkeley.edu/~storn/code.html#deb1.
"""

# Copyright (C) 2011, 2012, 2013, 2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
from builtins import range
__author__ = 'Benjamin Jonen <benjamin.jonen@bf.uzh.ch>'
__docformat__ = 'reStructuredText'

import numpy as np

from gc3libs.optimizer import EvolutionaryAlgorithm, populate
from gc3libs.utils import Enum

# in code one can use:
#  `strategies.DE_rand`
# or:
#  'DE_rand'
strategies = Enum(
    'DE_rand',
    'DE_local_to_best',
    'DE_best_with_jitter',
    'DE_rand_with_per_vector_dither',
    'DE_rand_with_per_generation_dither',
    'DE_rand_either_or_algorithm'
)


class DifferentialEvolutionAlgorithm(EvolutionaryAlgorithm):

    '''Differential Evolution Algorithm class.
    :class:`DifferentialEvolutionAlgorithm` explicitly allows for an another
    process to control the optimization. Driver classes can be found
    in `gc3libs.optimizer.drivers.py`:mod:.

    :param initial_pop: Initial population for the optimization.
      Value can be any sequence that can be passed to the `np.array()` constructor.
    :param str de_strategy: e.g. DE_rand_either_or_algorithm. Allowed are:
    :param float `de_step_size`: Differential Evolution step size.
    :param float `prob_crossover`: Probability new population draws will replace old members.
    :param bool exp_cross: Set True to use exponential crossover.
    :param int `itermax`: Maximum # of iterations.
    :param float `dx_conv_crit`: Abort optimization if all population members are within a certain distance to each other.
    :param float `y_conv_crit`: Declare convergence when the target function is below a `y_conv_crit`.
    :param fun `in_domain`: Optional function that implements nonlinear constraints.
    :param float `seed`: Seed to initialize NumPy's random number generator.
    :param obj `logger`: Configured logger to use.
    :param `after_update_opt_state`: List of Functions that are called at the end of
                `DifferentialEvolutionAlgorithm.after_update_opt_state`:meth:. Use this list
                to provide problem-specific printing and plotting routines. Examples can be found
                in `gc3libs.optimizer.extra`:mod:.

    The `de_strategy` value must be chosen from the
    `dif_evolution.strategies` enumeration.  Allowed values are
    (description of the strategies taken from http://www.icsi.berkeley.edu/~storn/DeMat.zip):

    1. ``'DE_rand'``: The classical version of DE.
    2. ``'DE_local_to_best'``: A version which has been used by quite a number of
            scientists. Attempts a balance between robustness # and fast convergence.
    3. ``'DE_best_with_jitter'``: Taylored for small population sizes and fast
            convergence. Dimensionality should not be too high.
    4. ``'DE_rand_with_per_vector_dither'``: Classical DE with dither to become even more robust.
    5. ``'DE_rand_with_per_generation_dither'``: Classical DE with dither to become even more robust.
                                           Choosing de_step_size = 0.3 is a good start here.
    6. ``'DE_rand_either_or_algorithm'``: Alternates between differential mutation and three-point- recombination.

    '''

    def __init__(self, initial_pop,
                 # DE-specific parameters
                 de_strategy='DE_rand', de_step_size=0.85, prob_crossover=1.0, exp_cross=False,
                 # converge-related parameters
                 itermax=100, dx_conv_crit=None, y_conv_crit=None,
                 # misc
                 in_domain=None, seed=None, logger=None, after_update_opt_state=[]):

        # Check input variables
        assert 0.0 <= prob_crossover <= 1.0, "prob_crossover should be from interval [0,1]"
        assert len(
            initial_pop) >= 5, "DifferentialEvolution requires at least 5 vectors in the population!"

        # initialize base class
        EvolutionaryAlgorithm.__init__(
            self,
            initial_pop,
            itermax, dx_conv_crit, y_conv_crit,
            logger, after_update_opt_state
        )
        # save parameters
        self.de_step_size = de_step_size
        self.prob_crossover = prob_crossover
        self.exp_cross = exp_cross
        self.de_strategy = de_strategy

        if not in_domain:
            self.in_domain = self._default_in_domain
        else:
            self.in_domain = in_domain

        # initialize NumPy's RNG
        np.random.seed(seed)

    def _default_in_domain(self, x):
        return np.array([True] * self.pop_size)

    def select(self, new_pop, new_vals):
        '''
        Perform a one-on-one battle by index, keeping the member
        with lowest corresponding value.
        '''
        ix_superior = new_vals < self.vals
        self.pop[ix_superior, :] = new_pop[ix_superior, :].copy()
        self.vals[ix_superior] = new_vals[ix_superior].copy()

    def evolve(self):
        '''
        Generates a new population fullfilling `in_domain`.

        :rtype: list of population members
        '''
        return populate(
            create_fn=(
                lambda: DifferentialEvolutionAlgorithm.evolve_fn(
                    self.pop,
                    self.prob_crossover,
                    self.de_step_size,
                    self.dim,
                    self.best_x,
                    self.de_strategy,
                    self.exp_cross)),
            in_domain=self.in_domain)

    @staticmethod
    def evolve_fn(
            population,
            prob_crossover,
            de_step_size,
            dim,
            best_iter,
            de_strategy,
            exp_cross):
        """
        Return new population, evolved according to `de_strategy`.

        :param population: Population generating offspring from.
        :param prob_crossover: Probability new population draws will replace old members.
        :param de_step_size: Differential Evolution step size.
        :param dim: Dimension of each population member.
        :param best_iter: Best population member of the current population.
        :param de_strategy: Differential Evolution strategy. See :class:`DifferentialEvolutionAlgorithm`.
        :param exp_cross bool: Set True to use exponential crossover.
        """

        assert de_strategy in strategies

        pop_size = len(population)

        # BJ: Need to add +1 in definition of ind otherwise there is one zero
        # index that leaves creates no shuffling.
        # index pointer array. e.g. [2, 1, 4, 3]
        ind = np.random.permutation(4) + 1
        rot = np.arange(
            0,
            pop_size,
            1)     # rotating index array (size pop_size)
        rt = np.zeros(pop_size)            # another rotating index array
        # index arrays
        a1 = np.random.permutation(pop_size)   # shuffle locations of vectors
        a2 = a1[
            (rot + ind[0]) %
            pop_size]  # rotate vector locations by ind[0] positions
        a3 = a2[(rot + ind[1]) % pop_size]
        a4 = a3[(rot + ind[2]) % pop_size]
        a5 = a4[(rot + ind[3]) % pop_size]

        pm1 = population[a1, :]  # shuffled population matrix 1
        pm2 = population[a2, :]  # shuffled population matrix 2
        pm3 = population[a3, :]  # shuffled population matrix 3
        pm4 = population[a4, :]  # shuffled population matrix 4
        pm5 = population[a5, :]  # shuffled population matrix 5

        # "best member" matrix
        bm = np.zeros((pop_size, dim))   # initialize FVr_bestmember  matrix
        # population filled with the best member
        for k in range(pop_size):
            bm[k, :] = best_iter                       # of the last iteration

        # mask for intermediate population
        # all random numbers < prob_crossover are 1, 0 otherwise
        mui = np.random.random_sample((pop_size, dim)) < prob_crossover

        if exp_cross:
            # rotating index array, i.e. [0, 1, 2, ..., dim]
            rotd = np.arange(dim)
            # rotating index array for exponential crossover
            rtd = np.zeros(dim, dtype=np.int)
            # Prepare intermediate population for indexing.
            mui = np.sort(mui.transpose(), axis=0)
            # Columns are pop members. Put all False indices in the first rows.
            for k in range(pop_size):
                n = int(np.floor(np.random.rand() * dim))
                if n > 0:
                    # Build actual rotation vector.
                    rtd = (rotd + n) % dim
                    # e.g. dim = 6, n = 2 -> [3,2,1,0,1,2]
                    # Rotate indices for kth population member by n.
                    mui[:, k] = mui[rtd, k]
            mui = mui.transpose()

        # inverse mask to mui (mpo + mui == <vector of 1's>)
        mpo = mui < 0.5

        if (de_strategy == 'DE_rand'):
            #origin = pm3
            ui = pm3 + de_step_size * (pm1 - pm2)   # differential variation
            ui = population * mpo + ui * mui          # crossover
        elif (de_strategy == 'DE_local_to_best'):
            #origin = population
            ui = population + de_step_size * \
                (bm - population) + de_step_size * (pm1 - pm2)
            ui = population * mpo + ui * mui
        elif (de_strategy == 'DE_best_with_jitter'):
            #origin = bm
            ui = bm + (pm1 - pm2) * ((1 - 0.9999) * \
                       np.random.random_sample((pop_size, dim)) + de_step_size)
            ui = population * mpo + ui * mui
        elif (de_strategy == 'DE_rand_with_per_vector_dither'):
            #origin = pm3
            f1 = (
                (1 -
                 de_step_size) *
                np.random.random_sample(
                    (pop_size,
                     1)) +
                de_step_size)
            for k in range(dim):
                pm5[:, k] = f1
            ui = pm3 + (pm1 - pm2) * pm5    # differential variation
            ui = population * mpo + ui * mui     # crossover
        elif (de_strategy == 'DE_rand_with_per_generation_dither'):
            #origin = pm3
            f1 = (
                (1 -
                 de_step_size) *
                np.random.random_sample() +
                de_step_size)
            ui = pm3 + (pm1 - pm2) * f1         # differential variation
            ui = population * mpo + ui * mui   # crossover
        elif (de_strategy == 'DE_rand_either_or_algorithm'):
            #origin = pm3
            # Pmu = 0.5
            if (np.random.random_sample() < 0.5):
                ui = pm3 + de_step_size * (pm1 - pm2)  # differential variation
            # use F-K-Rule: K = 0.5(F+1)
            else:
                ui = pm3 + 0.5 * (de_step_size + 1.0) * (pm1 + pm2 - 2 * pm3)
                ui = population * mpo + ui * mui     # crossover

        return ui

    # Adjustments for pickling
    def __getstate__(self):
        state = self.__dict__.copy()
        if 'logger' in list(state.keys()):
            del state['logger']
#        return state
        return None

    def __setstate__(self, state):
        self.__dict__ = state


# Variable changes from matlab implementation
# I_D -> dim
# I_NP -> pop_size
# FM_popold -> pop_old
# FVr_bestmem -> best
# FVr_bestmemit -> best_cur_iter
# I_nfeval -> n_fun_evals
# I_cur_iter -> cur_iter
# F_weight -> de_step_size
# F_CR -> prob_crossover
# I_itermax -> itermax
# F_VTR -> y_conv_crit
# I_strategy -> de_strategy
# I_plotting -> plotting
# lower_bds
# upper_bds
# dx_conv_crit
# verbosity
# FM_pm1 -> pm1
# FM_pm2 -> pm2 population matrix (pm)
# FM_pm3 -> pm3
# FM_pm4 -> pm4
# FM_pm5 -> pm5
# FM_bm  -> bm best member matrix
# FM_ui  -> ui ??
# FM_mui -> mui # mask for intermediate population
# FM_mpo -> mpo # mask for old population
# FVr_rot -> rot  # rotating index array (size I_NP)
# FVr_rotd -> rotd  # rotating index array (size I_D)
# FVr_rt -> rt  # another rotating index array
# FVr_rtd -> rtd # rotating ininstalldex array for exponential crossover
# FVr_a1 -> a1 # index array
# FVr_a2 -> a2 # index array
# FVr_a3 -> a3 # index array
# FVr_a4 -> a4 # index array
# FVr_a5 -> a5 # index array
# FVr_ind -> ind # index pointer array
# I_best_index -> best_ix
# S_vals -> vals
# S_bestval -> best_y
# S_bestvalit -> best_y_iter
# best -> best_x
# best_iter -> best_x_iter



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2017-2018,  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
from builtins import object
__docformat__ = 'reStructuredText'

import os
import shutil
import tempfile
from time import sleep

import pytest

import gc3libs.poller as plr
from gc3libs.utils import write_contents


# for readability
def _check_events(poller, path, expected):
    events = [event for event in poller.get_new_events()
              if event[0].path == path]
    assert len(events) == len(expected)
    for n, expected_event in enumerate(expected):
        url, actual_event = events[n]
        assert url.path == path
        assert actual_event == expected_event


class TestPollers(object):

    @pytest.fixture(autouse=True)
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        yield
        shutil.rmtree(self.tmpdir)


    def check_file_accesspoller(self):
        poller = plr.FilePoller(self.tmpdir)
        assert self.tmpdir in poller._watched

        # test create file
        fpath = os.path.join(self.tmpdir, 'foo')
        write_contents(fpath, 'test')
        _check_events(poller, fpath, ['created'])
        assert fpath in poller._watched

        # since the `mtime` field only has 1-second resolution, we
        # need to pause a bit to ensure the following
        # `.get_new_events()` are able to see any modification
        sleep(2)

        # no new events here
        _check_events(poller, fpath, [])

        # test modify file
        write_contents(fpath, 'test2')
        _check_events(poller, fpath, ['modified'])

        # test remove file
        os.remove(fpath)
        _check_events(poller, fpath, ['deleted'])
        assert fpath not in poller._watched


    def test_inotifypoller(self):
        poller = plr.INotifyPoller(self.tmpdir)

        # test create file
        fpath = os.path.join(self.tmpdir, 'foo')
        write_contents(fpath, 'test')
        _check_events(poller, fpath, ['created'])

        # no new events here
        _check_events(poller, fpath, [])

        # test modify file
        write_contents(fpath, 'test2')
        _check_events(poller, fpath, ['modified'])

        # test remove file
        os.remove(fpath)
        _check_events(poller, fpath, ['deleted'])


## main: run tests

if "__main__" == __name__:
    pytest.main(["-v", __file__])



#!/usr/bin/env python
#
#   gsaphir.py -- Front-end script for submitting multiple saphir jobs.
#
#   last modified: 28.09.2017 by Akos Dobay
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `saphir` jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gsaphir --help`` for program usage instructions.
"""

__version__ = 'development version 1.0 (SVN $Revision$)'
__author__ = 'Akos Dobay'
__docformat__ = 'reStructuredText'


import os
import glob
import decimal
import itertools
import tarfile
import tempfile

import gc3libs
from gc3libs import Application, Run
from gc3libs.cmdline import SessionBasedScript, executable_file, existing_file
from gc3libs.workflow import ChunkedParameterSweep
from gc3libs.utils import irange, parse_range, write_contents

if __name__ == '__main__':
    import gsaphir
    gsaphir.GSaphirScript().run()


class GSaphirApplication(Application):
    """
    Class wrapping the execution of a single `saphir` program.
    """

    application_name = "gsaphir"

    def __init__(self,
                 executable,
                 search_radius,
                 exclusion_zone,
                 scaling_factor,
                 intensity_threshold,
                 input_file,
                 input_image,
                 **extra_args):

        extra_args.setdefault('requested_cores', 1)

        self.executable = executable
        self.search_radius = search_radius
        self.exclusion_zone = exclusion_zone
        self.scaling_factor = scaling_factor
        self.intensity_threshold = intensity_threshold
        self.input_file = input_file
        self.input_image = input_image

        extra_args.setdefault('jobname',
                              "GSaphir:%i:%i:%i:%f:%s"
                              % (search_radius, exclusion_zone, scaling_factor, intensity_threshold,
                                 os.path.basename(input_image)))

        if 'output_dir' in extra_args:
            extra_args['output_dir'] = os.path.join(
                os.path.dirname(extra_args['output_dir']),
                extra_args['jobname'])

        executable_script = """#!/bin/sh

# exit immediately on error
set -e

echo "==> Working directory contents:"
ls -l

# Create data directory

mkdir -pv data

# Run the script

echo "==> Running 'saphir' ..."
./{saphir} -r {search_radius} -e {exclusion_zone} -c {scaling_factor} -t {intensity_threshold} -f {input_file} -s {input_image} -D data

# Compress all output files

#gzip data/*.txt
#gzip data/*.png

""".format(
    saphir=os.path.basename(executable),
    search_radius=int(search_radius),
    exclusion_zone=int(exclusion_zone),
    scaling_factor=int(scaling_factor),
    intensity_threshold=float(intensity_threshold),
    input_file=os.path.basename(input_file),
    input_image=os.path.basename(input_image),
)
        try:
            (fd, self.tmp_filename) = tempfile.mkstemp(prefix='gc3pie-gsaphir_')
            write_contents(self.tmp_filename, executable_script)
            os.chmod(self.tmp_filename, 0o755)
        except Exception as ex:
            gc3libs.log.error(
                "Error creating execution script: %s: %s",
                type(ex), ex)
            raise

        input_files = {
            # local abs path : remote file name
            executable: os.path.basename(executable),
            self.tmp_filename: 'gsaphir.sh',
            input_file: os.path.basename(input_file),
            input_image: os.path.basename(input_image),
        }

        output_files = {
            # download this remote file ...
            'data/dimension.txt':
                # ...into this local one
                'dimension.txt',
            # download this remote file ...
            # ??? how is the output image named?
            os.path.join('data', os.path.basename(self.input_image)):
                # ...into this local one
                ('er_{er}_sr_{sr}_sf_{sf}_it_{it}_{image}'
                 .format(
                     er=self.exclusion_zone,
                     sr=self.search_radius,
                     sf=self.scaling_factor,
                     it=(int((1 - self.intensity_threshold) * 100)),
                     image=os.path.basename(self.input_image)))
        }

        Application.__init__(
            self,
            ['./gsaphir.sh'],
            inputs=input_files,
            outputs=output_files,
            stdout="gsaphir.out",
            stderr="gsaphir.err",
            **extra_args)

    def terminated(self):
        """
        Remove temporary script file
        """
        try:
            os.remove(self.tmp_filename)
        except Exception as ex:
            # non-fatal error: log and ignore
            gc3libs.log.warning(
                "Failed removing temporary file %s: %s: %s ",
                self.tmp_filename, type(ex), ex)


class GSaphirTaskCollection(ChunkedParameterSweep):

    def __init__(self,
                 executable,
                 search_radius_range,
                 exclusion_zone_range,
                 scaling_factor_range,
                 intensity_threshold_range,
                 input_file,
                 input_image,
                 chunk_size,
                 **extra_args):

        self.executable = executable
        self.extra_args = extra_args
        self.input_file = input_file
        self.input_image = input_image

        self.combinations = list(
            itertools.product(
                irange(*search_radius_range),
                irange(*exclusion_zone_range),
                irange(*scaling_factor_range),
                irange(*intensity_threshold_range),
            )
        )

        ChunkedParameterSweep.__init__(
            self,
            0, # min value
            len(self.combinations), # max value, has to be at least 2
            1, # step
            chunk_size,
            **extra_args
        )

    def new_task(self, param, **extra):
        sr, ez, sf, it = self.combinations[param]
        return GSaphirApplication(
            self.executable,
            sr, ez, sf, it, self.input_file, self.input_image,
            **self.extra_args)


class GSaphirScript(SessionBasedScript):
    """
    Run multiple instances of `saphir` with the combination of
    supplied parameters.
    """
    version = __version__
    application_name = 'gsaphir'

    def setup_options(self):
        self.add_param("-a", "--search-radius", required=True,
                       help="In the form N[:END:STEP]. If only `N` is "
                       "supplied, will use only that value, otherwise "
                       "will use all the values in the range from `N` "
                       "to `END` (exclusive) using `STEP` increments")

        self.add_param("-e", "--exclusion-zone", required=True,
                       help="In the form N[:END:STEP]. If only `N` is "
                        "supplied, will use only that value, otherwise "
                        "will use all the values in the range from `N` "
                        "to `END` (exclusive) using `STEP` increments")

        self.add_param("-b", "--scaling-factor", required=True,
                       help="In the form N[:END:STEP]. If only `N` is "
                            "supplied, will use only that value, otherwise "
                            "will use all the values in the range from `N` "
                            "to `END` (exclusive) using `STEP` increments")

        self.add_param("-t", "--intensity-threshold", required=True,
                       help="In the form N[:END:STEP]. If only `N` is "
                            "supplied, will use only that value, otherwise "
                            "will use all the values in the range from `N` "
                            "to `END` (exclusive) using `STEP` increments")

        self.add_param("-f", "--source-file", type=existing_file,
                       help="Read the original dimension file")

        self.add_param("-I", "--source-image", type=existing_file,
                       help="Read the image source")

        self.add_param("-x", "--executable", type=executable_file,
                       default='saphir', metavar='FILE',
                       help="Path to the saphir executable.")

    def parse_args(self):
        if self.params.search_radius:
            self.params.search_radius_range = parse_range(self.params.search_radius)

        if self.params.exclusion_zone:
            self.params.exclusion_zone_range = parse_range(self.params.exclusion_zone)

        if self.params.scaling_factor:
            self.params.scaling_factor_range = parse_range(self.params.scaling_factor)

        if self.params.intensity_threshold:
            self.params.intensity_threshold_range = parse_range(self.params.intensity_threshold)

        self.params.executable = os.path.abspath(self.params.executable)

    def new_tasks(self, extra):
        return [
            GSaphirTaskCollection(
                self.params.executable,
                self.params.search_radius_range,
                self.params.exclusion_zone_range,
                self.params.scaling_factor_range,
                self.params.intensity_threshold_range,
                self.params.source_file,
                self.params.source_image,
                1000, # FIXME! make an option, do not hard-code
                **extra.copy()
            )
        ]
